library(RColorBrewer)
library(grDevices)

## TODO: Add the possibility to read the data directly from netCDF
## TODO: Get variable and system/obs names from dictionary
## TODO: Adapt to multi-model case

plot_data <- function(recipe,
		      archive,
		      data,
		      calibrated_data = NULL,
		      skill_metrics = NULL,
		      significance = F) {

  # Try to produce and save several basic plots.
  # recipe: the auto-s2s recipe as read by read_yaml()
  # archive: the auto-s2s archive as read by read_yaml()
  # data: list containing the hcst, obs and (optional) fcst s2dv_cube objects
  # calibrated_data: list containing the calibrated hcst and (optional) fcst
  # s2dv_cube objects
  # skill_metrics: list of arrays containing the computed skill metrics
  # significance: Bool. Whether to include significance dots where applicable

  outdir <- paste0(get_dir(recipe), "/plots/")
  dir.create(outdir, showWarnings = FALSE, recursive = TRUE)

  if ((is.null(skill_metrics)) && (is.null(calibrated_data)) && 
      (is.null(data$fcst))) {
    stop("The Visualization module has been called, but there is no data ",
	 "that can be plotted.")
  }
  if (!is.null(skill_metrics)) {
    plot_enscorr(recipe, archive, data$hcst, skill_metrics, outdir)
    plot_skill_scores(recipe, archive, data$hcst, skill_metrics, outdir,
		      significance)
    plot_scores(recipe, archive, data$hcst, skill_metrics, outdir)
  }
  if (!is.null(calibrated_data$fcst)) {
    plot_ensemble_mean(recipe, archive, calibrated_data$fcst, outdir)
  } else if (!is.null(data$fcst)) {
    warning("Only the uncalibrated forecast was provided. Using this data ",
	    "to plot the forecast ensemble mean.")
    plot_ensemble_mean(recipe, archive, data$fcst, outdir)
  }
  print("##### PLOTS SAVED TO OUTPUT DIRECTORY #####")
}


plot_enscorr <- function(recipe, archive, data_cube, skill_metrics, outdir) {
  ## TODO: Is this function even necessary...?

  latitude <- data_cube$lat[1:length(data_cube$lat)]
  longitude <- data_cube$lon[1:length(data_cube$lon)]
  system_name <- archive$System[[recipe$Analysis$Datasets$System$name]]$name

  # Define plot characteristics
  brks <- seq(-1, 1, by = 0.1)
  col2 <- grDevices::hcl.colors(length(brks) - 1, "RdBu", rev = TRUE)
  # color <- colorRampPalette(col2)(length(brks) - 1)
  options(bitmapType = "cairo")

  enscorr_names <- c("enscorr", "enscorr_specs")

  for (name in enscorr_names) {
    if (name %in% names(skill_metrics)) {
      enscorr <- skill_metrics[[name]]
      enscorr <- Reorder(enscorr, c("time", "longitude", "latitude"))
      outfile <- paste0(outdir, name, ".png")
      ## TODO: Put plot info in the titles (startdate, month)
      toptitle <- paste0("Ensemble Mean Correlation - ",
			 data_cube$Variable$varName, " - ",
			 system_name)
      months <- unique(lubridate::month(data$hcst$Dates$start,
				        label = T, abb = F))
      titles <- as.vector(months)
      # Plot
      PlotLayout(PlotEquiMap, c('longitude', 'latitude'),
                 enscorr, longitude, latitude, 
		 toptitle = toptitle,
		 titles = titles,
		 filled.continents=F, 
                 brks = brks,
                 #rxow_titles=row_titles, 
                 cols = col2, 
                 fileout = outfile,
	         bar_label_digits = 1)
    }
  }
}

plot_skill_scores <- function(recipe, archive, data_cube, skill_metrics, outdir,
			      significance = F) {
  
  latitude <- data_cube$lat[1:length(data_cube$lat)]
  longitude <- data_cube$lon[1:length(data_cube$lon)]
  system_name <- archive$System[[recipe$Analysis$Datasets$System$name]]$name
  
  # Define plot characteristics
  brks <- seq(-1, 1, by = 0.1)
  col2 <- grDevices::hcl.colors(length(brks) - 1, "RdBu", rev = TRUE)
  # color <- colorRampPalette(col2)(length(brks) - 1)
  options(bitmapType = "cairo")

  skill_scores <- c("rpss", "bss90", "bss10", "frpss", "crpss", "mean_bias_ss")

  for (name in skill_scores) {
    if (name %in% names(skill_metrics)) {
      skill <- skill_metrics[[name]]
      skill <- Reorder(skill, c("time", "longitude", "latitude"))
      if (significance) {
        significance_name <- paste0(name, "_significance")
        skill_significance <- skill_metrics[[significance_name]]
        skill_significance <- Reorder(skill_significance, c("time", 
							    "longitude",
							    "latitude"))
      } else {
	skill_significance <- NULL
      }
      # Define output file name and titles
      outfile <- paste0(outdir, name, ".png")
      ## TODO: Put plot info in the titles (startdate, month)
      toptitle <- paste0(toupper(name), " - ", data_cube$Variable$varName,
			 " - ", system_name)
      months <- unique(lubridate::month(data$hcst$Dates$start,
				        label = T, abb = F))
      titles <- as.vector(months)
      # Plot
      PlotLayout(PlotEquiMap, c('longitude', 'latitude'),
                 skill, longitude, latitude,
		 dots = skill_significance,
		 toptitle = toptitle,
		 titles = titles,
		 filled.continents=F, 
                 brks = brks,
                 #rxow_titles=row_titles, 
                 cols = col2,
                 fileout = outfile,
	         bar_label_digits = 1)
    }
  }
}

plot_scores <- function(recipe, archive, data_cube, skill_metrics, outdir) {
  
  latitude <- data_cube$lat[1:length(data_cube$lat)]
  longitude <- data_cube$lon[1:length(data_cube$lon)]
  system_name <- archive$System[[recipe$Analysis$Datasets$System$name]]$name

  # Define plot characteristics
  brks <- seq(0, 1, by = 0.1)
  col2 <- grDevices::hcl.colors(length(brks) - 1, "Reds")
  # color <- colorRampPalette(col2)(length(brks) - 1)
  options(bitmapType = "cairo")

  scores <- c("rps", "frps", "crps")

  for (name in scores) {
    if (name %in% names(skill_metrics)) {
      skill <- skill_metrics[[name]]
      skill <- Reorder(skill, c("time", "longitude", "latitude"))
      # Define name of output file and titles
      outfile <- paste0(outdir, name, ".png")
      toptitle <- paste0(toupper(name), " - ", data_cube$Variable$varName,
			 " - ", system_name)
      months <- unique(lubridate::month(data$hcst$Dates$start,
				        label = T, abb = F))
      titles <- as.vector(months)
      # Plot
      PlotLayout(PlotEquiMap, c('longitude', 'latitude'),
                 skill, longitude, latitude,
		 toptitle = toptitle,
		 titles = titles,
		 filled.continents=F, 
                 brks = brks,
                 #rxow_titles=row_titles, 
                 cols = col2, 
                 fileout = outfile,
	         bar_label_digits = 1)
    }
  }
}

plot_ensemble_mean <- function(recipe, archive, data_cube, outdir) {

  latitude <- data_cube$lat[1:length(data_cube$lat)]
  longitude <- data_cube$lon[1:length(data_cube$lon)]
  system_name <- archive$System[[recipe$Analysis$Datasets$System$name]]$name
  variable <- recipe$Analysis$Variables$name

  # Compute ensemble mean
  ensemble_mean <- s2dv::MeanDims(data_cube$data, 'ensemble')
  # Drop extra dims, add time dim if missing:
  ensemble_mean <- drop(ensemble_mean)
  if (!("time" %in% names(dim(ensemble_mean)))) {
    dim(ensemble_mean) <- c("time" = 1, dim(ensemble_mean))
  }
  ensemble_mean <- Reorder(ensemble_mean, c("time", "longitude", "latitude"))
  ## TODO: Redefine column colors, possibly depending on variable
  if (variable == 'prlr') {
    palette = "BrBG"
    rev = F
  } else {
    palette = "RdBu"
    rev = T
  }

  brks <- pretty(range(ensemble_mean, na.rm = T), n = 10, min.n = 8)
  col2 <- grDevices::hcl.colors(length(brks) - 1, palette, rev = rev)
  # color <- colorRampPalette(col2)(length(brks) - 1)
  options(bitmapType = "cairo")

  # Define name of output file and titles
  outfile <- paste0(outdir, "forecast_ensemble_mean.png")
  toptitle <- paste0("Ensemble Mean - ", variable, " - ", system_name)
  months <- unique(lubridate::month(data$hcst$Dates$start,
				        label = T, abb = F))
  titles <- as.vector(months)
  # Plots
  PlotLayout(PlotEquiMap, c('longitude', 'latitude'),
	     ensemble_mean, longitude, latitude,
             filled.continents = F,
	     toptitle = toptitle,
	     titles = titles,
	     cols = col2,
	     brks = brks,
	     fileout = outfile,
             bar_label_digits=4)
}
